!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! nbody_particle_star.f90 !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      MODULE nbody_particle_star

          INTEGER,PARAMETER :: memory_success = 0

          INTEGER,PARAMETER :: default_string_length = 1024

          INTEGER,PARAMETER :: minimum_file_unit =  0
          INTEGER,PARAMETER :: maximum_file_unit = 99
          INTEGER,PARAMETER :: stdin_file_unit   =  5
          INTEGER,PARAMETER :: stdout_file_unit  =  6
          INTEGER,PARAMETER :: stderr_file_unit  =  0

          INTEGER,PARAMETER :: io_success        =  0
          INTEGER,PARAMETER :: io_end_of_file    = -1

          REAL,PARAMETER :: gravitational_constant = 6.6726e-11 !! N(m/kg)^2

          REAL,PARAMETER :: default_domain_size_x  = 1.0e+18 !! m
          REAL,PARAMETER :: default_domain_size_y  = 1.0e+18 !! m
          REAL,PARAMETER :: default_domain_size_z  = 1.0e+18 !! m
          REAL,PARAMETER :: default_mass_maximum   = 1.0e+18 !! kg
          REAL,PARAMETER :: default_time_interval  = 1.0e+18 !! s

          INTEGER,PARAMETER :: default_number_of_particles = 1000

          INTEGER,PARAMETER :: default_number_of_timesteps =  100
          INTEGER,PARAMETER :: default_timesteps_between_outputs =  &
     &                             default_number_of_timesteps

          INTEGER,PARAMETER :: default_random_seed = 12345

          CHARACTER (LEN = default_string_length),PARAMETER ::  &
     &        default_base_filename = "nbody"

          TYPE Particle
              REAL :: position_x !! m
              REAL :: position_y !! m
              REAL :: position_z !! m
              REAL :: velocity_x !! m/s
              REAL :: velocity_y !! m/s
              REAL :: velocity_z !! m/s
              REAL :: force_x    !! N
              REAL :: force_y    !! N
              REAL :: force_z    !! N
              REAL :: mass       !! kg
          END TYPE Particle

          TYPE(Particle),DIMENSION(:),ALLOCATABLE :: particle_array

          CHARACTER (LEN = default_string_length) ::  &
     &        base_filename = default_base_filename

          REAL :: domain_size_x = default_domain_size_x
          REAL :: domain_size_y = default_domain_size_y
          REAL :: domain_size_z = default_domain_size_z
          REAL :: mass_maximum  = default_mass_maximum
          REAL :: time_interval = default_time_interval

          INTEGER :: number_of_particles = default_number_of_particles

          INTEGER :: number_of_timesteps = default_number_of_timesteps
          INTEGER :: timesteps_between_outputs =  &
     &                   default_timesteps_between_outputs

          INTEGER :: random_seed = default_random_seed

          LOGICAL :: debug = .TRUE.

      CONTAINS !! nbody_particle_star

!........ Input the run arguments from standard input.

          SUBROUTINE Particle_input_arguments ()
              IMPLICIT NONE

              number_of_particles       = default_number_of_particles
              domain_size_x             = default_domain_size_x
              domain_size_y             = default_domain_size_y
              domain_size_z             = default_domain_size_z
              time_interval             = default_time_interval
              number_of_timesteps       = default_number_of_timesteps
              timesteps_between_outputs =   &
     &            default_timesteps_between_outputs
              random_seed               = default_random_seed
              mass_maximum              = default_mass_maximum

              READ *, number_of_particles

              IF (number_of_particles < 1) THEN

                  WRITE (0,*) "ERROR: cannot have ",  &
     &                number_of_particles, "particles!"

                  STOP

              END IF !! (number_of_particles < 1)

              IF (number_of_particles == 1) THEN

                  WRITE (0,*) "There is only one particle, ",  &
     &                "therefore no forces."

                  STOP

              END IF !! (number_of_particles == 1)

              READ *, domain_size_x

              IF (domain_size_x <= 0.0) THEN

                  WRITE (0,*) "ERROR: cannot have a domain whose ",  &
     &                "X dimension has length ", domain_size_x, "!"

                  STOP

              END IF !! (domain_size_x <= 0.0)

              READ *, domain_size_y

              IF (domain_size_y <= 0.0) THEN

                  WRITE (0,*) "ERROR: cannot have a domain whose ",  &
     &                "Y dimension has length ", domain_size_y, "!"

                  STOP

              END IF !! (domain_size_y <= 0.0)

              READ *, domain_size_z

              IF (domain_size_z <= 0.0) THEN

                  WRITE (0,*) "ERROR: cannot have a domain whose ",  &
     &                "Z dimension has length ", domain_size_z, "!"

                  STOP

              END IF !! (domain_size_z <= 0.0)

              READ *, time_interval

              IF (time_interval <= 0.0) THEN

                  WRITE (0,*) "ERROR: cannot have a time interval ",  &
     &                "of ", time_interval, "!"

                  STOP

              END IF !! (time_interval <= 0.0)

              READ *, number_of_timesteps

              IF (number_of_timesteps <= 0) THEN

                  WRITE (0,*) "ERROR: cannot have ", &
     &                number_of_timesteps, " timesteps!"

                  STOP

              END IF !! (number_of_timesteps <= 0)

              READ *, timesteps_between_outputs

              IF (timesteps_between_outputs <= 0) THEN

                  WRITE (0,*) "ERROR: cannot have ", &
     &                timesteps_between_outputs,     &
     &                " timesteps between outputs!"

                  STOP

              END IF !! (timesteps_between_outputs <= 0)

              READ *, random_seed

              IF (random_seed < 0) THEN

                  WRITE (0,*) "ERROR: cannot have a random seed of ", &
     &                random_seed, "!"

                  STOP

              END IF !! (random_seed < 0)

              READ *, mass_maximum

              IF (mass_maximum <= 0.0) THEN

                  WRITE (0,*) "ERROR: cannot have ",  &
     &                "a maximum mass of ", mass_maximum, "!"

                  STOP

              END IF !! (mass_maximum <= 0.0)

              READ *, base_filename

          END SUBROUTINE Particle_input_arguments

!........ Construct the particle.

          SUBROUTINE Particle_construct (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(OUT) :: this_particle

              CALL Particle_clear(this_particle)

          END SUBROUTINE Particle_construct

!........ Clear the particle's data.

          SUBROUTINE Particle_clear (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(OUT) :: this_particle

              this_particle % position_x = 0.0
              this_particle % position_y = 0.0
              this_particle % position_z = 0.0
              this_particle % velocity_x = 0.0
              this_particle % velocity_y = 0.0
              this_particle % velocity_z = 0.0
              this_particle % force_x    = 0.0
              this_particle % force_y    = 0.0
              this_particle % force_z    = 0.0
              this_particle % mass       = 0.0

          END SUBROUTINE Particle_clear

!........ Initialize the particle by setting its data randomly.

          SUBROUTINE Particle_set_position_randomly (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(INOUT) :: this_particle

              REAL :: random_value

              CALL RANDOM_NUMBER(random_value)
              this_particle % position_x =  &
     &            domain_size_x * random_value

              CALL RANDOM_NUMBER(random_value)
              this_particle % position_y =  &
     &            domain_size_y * random_value

              CALL RANDOM_NUMBER(random_value)
              this_particle % position_z =  &
     &            domain_size_z * random_value

          END SUBROUTINE Particle_set_position_randomly

!........ Initialize the particle by setting its data randomly.

          SUBROUTINE Particle_initialize_randomly (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(INOUT) :: this_particle

              REAL :: random_value

              CALL Particle_clear(this_particle)

              CALL Particle_set_position_randomly(this_particle)

              CALL RANDOM_NUMBER(random_value)
              this_particle % mass =  &
     &            mass_maximum * random_value

          END SUBROUTINE Particle_initialize_randomly

!........ Initialize the particle.

          SUBROUTINE Particle_initialize (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(INOUT) :: this_particle

              CALL Particle_initialize_randomly(this_particle)

          END SUBROUTINE Particle_initialize


!........ Calculate the force between the two particles.
!........
!........ Gregory R. Andrews, "Foundations of Multithreaded, Parallel,
!........ and Distributed Programming," Addison-Wesley, 2000, p. 554-6.

          SUBROUTINE Particle_pair_calculate_force (      &
     &                   this_particle1, this_particle2,  &
     &                   force_x, force_y, force_z)
              IMPLICIT NONE

              TYPE(Particle),INTENT(INOUT) :: this_particle1
              TYPE(Particle),INTENT(INOUT) :: this_particle2
              REAL,INTENT(OUT) :: force_x, force_y, force_z

              REAL :: difference_x, difference_y, difference_z
              REAL :: distance_squared, distance
              REAL :: force_magnitude

              difference_x =  &
     &            this_particle2 % position_x - this_particle1 % position_x
              difference_y =  &
     &            this_particle2 % position_y - this_particle1 % position_y
              difference_z =  &
     &            this_particle2 % position_z - this_particle1 % position_z

              distance_squared = difference_x * difference_x +  &
     &                           difference_y * difference_y +  &
     &                           difference_z * difference_z

              distance = SQRT(distance_squared)

              force_magnitude =                                &
     &            gravitational_constant *                     &
     &            this_particle1 % mass * this_particle2 % mass /  &
     &            distance_squared

              force_x = (force_magnitude / distance) * difference_x
              force_y = (force_magnitude / distance) * difference_y
              force_z = (force_magnitude / distance) * difference_z

          END SUBROUTINE Particle_pair_calculate_force

!........ Move the particle based on the force applied to it.
!........ 
!........ Gregory R. Andrews, "Foundations of Multithreaded, Parallel,
!........ and Distributed Programming," Addison-Wesley, 2000, p. 554-6.

          SUBROUTINE Particle_move (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(INOUT) :: this_particle

              REAL :: velocity_change_x
              REAL :: velocity_change_y
              REAL :: velocity_change_z
              REAL :: position_change_x
              REAL :: position_change_y
              REAL :: position_change_z

              velocity_change_x = this_particle % force_x *  &
     &                            (time_interval / this_particle % mass)
              velocity_change_y = this_particle % force_y *  &
     &                            (time_interval / this_particle % mass)
              velocity_change_z = this_particle % force_z *  &
     &                            (time_interval / this_particle % mass)

              position_change_x =  &
     &            this_particle % velocity_x +  &
     &            velocity_change_x * (0.5 * time_interval)
              position_change_y =  &
     &            this_particle % velocity_y +  &
     &            velocity_change_y * (0.5 * time_interval)
              position_change_z =  &
     &            this_particle % velocity_z +  &
     &            velocity_change_z * (0.5 * time_interval)

              this_particle % velocity_x =  &
     &            this_particle % velocity_x + velocity_change_x
              this_particle % velocity_y =  &
     &            this_particle % velocity_y + velocity_change_y
              this_particle % velocity_z =  &
     &            this_particle % velocity_z + velocity_change_z

              this_particle % position_x =  &
     &            this_particle % position_x + position_change_x
              this_particle % position_y =  &
     &            this_particle % position_y + position_change_y
              this_particle % position_z =  &
     &            this_particle % position_z + position_change_z

          END SUBROUTINE Particle_move

          SUBROUTINE Particle_output (output_file_unit, this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(IN) :: this_particle
              INTEGER,INTENT(IN) :: output_file_unit

              IF ((output_file_unit <  minimum_file_unit) .OR.  &
     &            (output_file_unit >  maximum_file_unit) .OR.  &
     &            (output_file_unit == stdin_file_unit)) THEN

                  WRITE (0,*) "ERROR: cannot write to file unit ",  &
     &                output_file_unit, "!"

                  STOP

              END IF !! ((output_file_unit <  minimum_file_unit) .OR. ...)

              WRITE (output_file_unit,  &
     &    "(E14.6,A,E14.6,A,E14.6,A,E14.6,A,E14.6,A,E14.6,A,E14.6)")  &
     &            this_particle % position_x, " ",  &
     &            this_particle % position_y, " ",  &
     &            this_particle % position_z, " ",  &
     &            this_particle % velocity_x, " ",  &
     &            this_particle % velocity_y, " ",  &
     &            this_particle % velocity_z, " ",  &
     &            this_particle % mass

          END SUBROUTINE Particle_output

!........ Destroy the particle.

          SUBROUTINE Particle_destroy (this_particle)
              IMPLICIT NONE

              TYPE(Particle),INTENT(INOUT) :: this_particle

              CALL Particle_clear(this_particle)

          END SUBROUTINE Particle_destroy

!........ Check the array of particles.

          SUBROUTINE Particle_array_check (action, routine)
              IMPLICIT NONE

              CHARACTER (LEN = *),INTENT(IN) :: action, routine

              IF (number_of_particles < 0) THEN

                  WRITE (0,*) "ERROR: illegal number of particles ",  &
     &                number_of_particles

                  IF (LEN_TRIM(action) < 1) THEN

                      WRITE (0,*) "  to perform an unknown action on"

                  ELSE   !! (LEN_TRIM(action) < 1)

                      WRITE (0,*) "  to ", TRIM(action)

                  END IF !! (LEN_TRIM(action) < 1)...ELSE

                  IF (LEN_TRIM(routine) < 1) THEN

                      WRITE (0,*) "  in an unknown routine!"

                  ELSE   !! (LEN_TRIM(routine) < 1)

                      WRITE (0,*) "  in ", TRIM(routine), "!"

                  END IF !! (LEN_TRIM(routine) < 1)...ELSE

                  STOP

              END IF !! (number_of_particles < 0)

              IF (number_of_particles == 0) THEN

                  IF (.NOT. ALLOCATED(particle_array)) THEN

                      RETURN

                  END IF !! (.NOT. ALLOCATED(particle_array))

                  IF (LEN_TRIM(action) < 1) THEN

                      WRITE (0,*) "  ERROR: cannot ",  &
     &                    "perform an unknown action on"

                  ELSE   !! (LEN_TRIM(action) < 1)

                      WRITE (0,*) "  ERROR: cannot ", TRIM(action)

                  END IF !! (LEN_TRIM(action) < 1)...ELSE

                  WRITE (0,*) "  an existing particle array ",  &
     &                "of length 0"

                  IF (LEN_TRIM(routine) < 1) THEN

                      WRITE (0,*) "  in an unknown routine!"

                  ELSE   !! (LEN_TRIM(routine) < 1)

                      WRITE (0,*) "  in ", TRIM(routine), "!"

                  END IF !! (LEN_TRIM(routine) < 1)...ELSE

                  STOP

              END IF !! (number_of_particles == 0)

              IF (.NOT. ALLOCATED(particle_array)) THEN

                  IF (LEN_TRIM(action) < 1) THEN

                      WRITE (0,*) "  ERROR: cannot ",  &
     &                    "perform an unknown action on"

                  ELSE   !! (LEN_TRIM(action) < 1)

                      WRITE (0,*) "  ERROR: cannot ", TRIM(action)

                  END IF !! (LEN_TRIM(action) < 1)...ELSE

                  WRITE (0,*) "  an nonexistent particle array ",  &
     &                "of length ", number_of_particles

                  IF (LEN_TRIM(routine) < 1) THEN

                      WRITE (0,*) "  in an unknown routine!"

                  ELSE   !! (LEN_TRIM(routine) < 1)

                      WRITE (0,*) "  in ", TRIM(routine), "!"

                  END IF !! (LEN_TRIM(routine) < 1)...ELSE

                  STOP

              END IF !! (.NOT. ALLOCATED(particle_array))

          END SUBROUTINE Particle_array_check

!........ Construct the array of particles.

          SUBROUTINE Particle_array_construct ()
              IMPLICIT NONE

              INTEGER :: part

              CALL Particle_array_allocate()

              DO part = 1, number_of_particles

                  CALL Particle_construct(particle_array(part))

              END DO !! part

          END SUBROUTINE Particle_array_construct

!........ Allocate the array of particles.

          SUBROUTINE Particle_array_allocate ()

              INTEGER :: memory_status
              INTEGER :: part

              IF (number_of_particles < 0) THEN

                  WRITE (0,*) "ERROR: illegal number of particles ",  &
     &                number_of_particles, " to allocate!"

                  WRITE (0,*) "  in Particle_array_allocate"

                  STOP

              END IF !! (number_of_particles < 0)

              IF (number_of_particles == 0) THEN

                  RETURN

              END IF !! (number_of_particles == 0)

              ALLOCATE(particle_array(1:number_of_particles),  &
     &                 STAT=memory_status)

              IF (memory_status /= memory_success) THEN

                  WRITE (0,*) "ERROR: cannot allocate ",  &
     &                "a particle array of ",             &
                      number_of_particles, "particles"
                  WRITE (0,*) "  in Particle_array_allocate!"

                  STOP

              END IF !! (memory_status /= memory_success)

          END SUBROUTINE Particle_array_allocate

!........ Initialize the array of particles.

          SUBROUTINE Particle_array_initialize ()
              IMPLICIT NONE

              CALL Particle_array_initialize_randomly()

          END SUBROUTINE Particle_array_initialize


!........ Initialize the array of particles by setting its data randomly.

          SUBROUTINE Particle_array_initialize_randomly ()
              IMPLICIT NONE

              INTEGER :: part

              DO part = 1, number_of_particles

                  CALL Particle_initialize_randomly(  &
     &                     particle_array(part))

              END DO !! part

          END SUBROUTINE Particle_array_initialize_randomly

!........ Calculate the forces between every pair of particles.

          SUBROUTINE Particle_array_calculate_forces ()
              IMPLICIT NONE

              REAL    :: particle_pair_force_x
              REAL    :: particle_pair_force_y
              REAL    :: particle_pair_force_z
              INTEGER :: part1, part2

              IF (number_of_particles <= 1) THEN

                  RETURN

              END IF !! (number_of_particles <= 1)

              particle_array(1:number_of_particles) % force_x = 0.0
              particle_array(1:number_of_particles) % force_y = 0.0
              particle_array(1:number_of_particles) % force_z = 0.0

              DO part1 = 1, number_of_particles

                  DO part2 = 1, number_of_particles

                      IF (part1 /= part2) THEN

                          CALL Particle_pair_calculate_force(  &
     &                             particle_array(part1),      &
     &                             particle_array(part2),      &
     &                             particle_pair_force_x,      &
     &                             particle_pair_force_y,      &
     &                             particle_pair_force_z)

                          particle_array(part1) % force_x =      &
     &                        particle_array(part1) % force_x +  &
     &                        particle_pair_force_x

                          particle_array(part1) % force_y =      &
     &                        particle_array(part1) % force_y +  &
     &                        particle_pair_force_y

                          particle_array(part1) % force_z =      &
     &                        particle_array(part1) % force_z +  &
     &                        particle_pair_force_z

                      END IF !! (part1 /= part2)

                  END DO !! part2

              END DO !! part1

          END SUBROUTINE Particle_array_calculate_forces

!........ Move every particle in the array of particles.

          SUBROUTINE Particle_array_move ()
              IMPLICIT NONE

              INTEGER :: part

              IF (number_of_particles <= 1) THEN

                  RETURN

              END IF !! (number_of_particles <= 1)

              DO part = 1, number_of_particles

                  CALL Particle_move(particle_array(part))

              END DO !! part

          END SUBROUTINE Particle_array_move

          SUBROUTINE Particle_array_output (base_filename, timestep)
              IMPLICIT NONE

              CHARACTER (LEN = *),INTENT(IN) :: base_filename
              INTEGER,INTENT(IN) :: timestep

              INTEGER,PARAMETER :: default_output_file_unit = 11

              CHARACTER (LEN = default_string_length) :: filename
              INTEGER :: output_file_unit
              INTEGER :: io_status
              INTEGER :: part

              IF (LEN_TRIM(base_filename) < 1) THEN

                  output_file_unit = stdout_file_unit

              ELSE   !! (LEN_TRIM(base_filename) < 1)

                  output_file_unit = default_output_file_unit

      WRITE (0,*) "base_filename = ", TRIM(base_filename)
      WRITE (0,*) "timestep = ", timestep

                  WRITE (filename, "(A,A,I8.8,A)")  &
     &                TRIM(base_filename), "_", timestep, ".txt"

      WRITE (0,*) "filename = ", TRIM(filename)

                  OPEN(UNIT=output_file_unit, FILE=TRIM(filename),  &
     &                 IOSTAT=io_status)

                  IF (io_status /= io_success) THEN

                      WRITE (0,*) "ERROR: cannot open the output file"
                      WRITE (0,*) "  named ", TRIM(filename)
                      WRITE (0,*) "  in Particle_array_output"

                      STOP

                  END IF !! (io_status /= io_success)

              END IF !! (LEN_TRIM(base_filename) < 1)...ELSE

              WRITE (output_file_unit,  &
     & "(I8.8,A,I8.8,A,I8.8,A,G14.8,A,G14.8,A,G14.8,A,G14.8,A,G14.8,A,I8.8)") &
     &            number_of_particles,       " ",  &
     &            number_of_timesteps,       " ",  &
     &            timesteps_between_outputs, " ",  &
     &            domain_size_x,             " ",  &
     &            domain_size_y,             " ",  &
     &            domain_size_z,             " ",  &
     &            mass_maximum,              " ",  &
     &            time_interval,             " ",  &
                  random_seed

              WRITE (output_file_unit, "(I8.8)") timestep

              DO part = 1, number_of_particles

                  CALL Particle_output(output_file_unit,  &
     &                                 particle_array(part))

              END DO !! part

              IF (output_file_unit /= stdout_file_unit) THEN

                  CLOSE(UNIT=output_file_unit, IOSTAT=io_status)

                  IF (io_status /= io_success) THEN

                      WRITE (0,*) "ERROR: cannot close the output file"
                      WRITE (0,*) "  named ", TRIM(filename)
                      WRITE (0,*) "  in Particle_array_output"

                      STOP

                  END IF !! (io_status /= io_success)

              END IF !! (output_file_unit /= stdout_file_unit)

          END SUBROUTINE Particle_array_output

!........ Destroy the array of particles.

          SUBROUTINE Particle_array_destroy ()
              IMPLICIT NONE

              CALL Particle_array_deallocate()

              number_of_particles = 0

          END SUBROUTINE Particle_array_destroy

!........ Deallocate the array of particles.

          SUBROUTINE Particle_array_deallocate ()

              INTEGER :: memory_status

              IF (.NOT. ALLOCATED(particle_array)) THEN

                  RETURN

              END IF !! (.NOT. ALLOCATED(particle_array))

              DEALLOCATE(particle_array, STAT=memory_status)

              IF (memory_status /= memory_success) THEN

                  WRITE (0,*) "ERROR: cannot deallocate ",  &
     &                "a particle array of ",               &
                      number_of_particles, "particles"
                  WRITE (0,*) "  in Particle_array_allocate!"

                  STOP

              END IF !! (memory_status /= memory_success)

          END SUBROUTINE Particle_array_deallocate

      END MODULE nbody_particle_star

